////////////////////////////////////////////////////////////////////////////////
// TSplitterControl class - Oxsoft Heart V5.0
//
// Class that allows to resize two panels (or so) which are separated by a kind
// of splitter
//
// by Alan GARNY
//    gry@physiol.ox.ac.uk
//    http://www.physiol.ox.ac.uk/~gry
//
//  Copyright 1996-97
////////////////////////////////////////////////////////////////////////////////
// Usage:
//     - Include the header into the one of your form, i.e.:
//          --------- CUT HERE ---------
//          #include "SplitterControl.h"
//          --------- CUT HERE ---------
//     - Declare a split control within the Private clause of the form, i.e.:
//          --------- CUT HERE ---------
//          TSplitterControl *SplitterControl;
//          --------- CUT HERE ---------
//     - Create the split control within the constructor of the form, i.e.:
//          --------- CUT HERE ---------
//          SplitterControl = new TSplitterControl(this);
//          --------- CUT HERE ---------
//     - Remove the split control within the destructor of the form, i.e.:
//          --------- CUT HERE ---------
//          delete SplitterControl;
//          --------- CUT HERE ---------
//     - Catch the "OnMouseDown" event of the splitter and start the splitting,
//       i.e.:
//          --------- CUT HERE ---------
//          if ((Button == mbLeft) && (Shift.Contains(ssLeft)))
//             SplitterControl->OnMouseDown(X, Y,
//                                          SplitterControl,
//                                          TargetControl, OtherTargetControl,
//                                          MinSize, MaxSize, ResizeVertically);
//          --------- CUT HERE ---------
//     - Catch the "OnMouseMove" event of the splitter and update the splitting,
//       i.e.:
//          --------- CUT HERE ---------
//          SplitterControl->OnMouseMove(X, Y);
//          --------- CUT HERE ---------
//     - Catch the "OnMouseUp" event of the splitter and end the splitting,
//       i.e.:
//          --------- CUT HERE ---------
//          SplitterControl->OnMouseUp(X, Y);
//          --------- CUT HERE ---------
////////////////////////////////////////////////////////////////////////////////
// Origin: Borland
//
// Improvements:
//    - Bug fixes,
//    - Better handling of the drawing of the splitter,
//    - New features (minimum/maximum size for instance).
////////////////////////////////////////////////////////////////////////////////
// Date of Creation: 23/04/97
//
// Modifications: (model: [<initials>, dd/mm/yy] <what has been done>)
////////////////////////////////////////////////////////////////////////////////

#include <vcl\vcl.h>
#pragma hdrstop

////////////////////////////////////////////////////////////////////////////////

#include "SplitterControl.h"

////////////////////////////////////////////////////////////////////////////////
// Constants

const int _NO_SIZE_ = -1;

////////////////////////////////////////////////////////////////////////////////
// TSplitterControl: constructor

__fastcall TSplitterControl::TSplitterControl(TForm *aForm)
   : Form(aForm),
     OrigPos(_NO_SIZE_)
{
}

////////////////////////////////////////////////////////////////////////////////
// OnMouseDown:

void __fastcall TSplitterControl::OnMouseDown(const int &aX, const int &aY,
                                              TControl *aSplitterControl,
                                              TControl *aTargetControl,
                                              TControl *aOtherTargetControl,
                                              const int &aMinSize,
                                              const int &aMaxSize,
                                              const bool &aResizeVertically)
{
   SplitterControl  = aSplitterControl;
   TargetControl    = aTargetControl;
   ResizeVertically = aResizeVertically;

   if ((aMinSize == _NO_SIZE_) || (aMinSize < 0))
      MinSize = 0;   // Default minimum size
   else
      MinSize = aMinSize;

   if ((aMaxSize == _NO_SIZE_) || (aMaxSize <= 0))
   {
      // Compute the default maximum size
      // Note: we need to substract 1 otherwise we get a strange behaviour
      //       (since the other target control has its "Align" property set to
      //       "alClient")

      if (ResizeVertically)
         MaxSize = aTargetControl->Height + aOtherTargetControl->Height - 1;
      else
         MaxSize = aTargetControl->Width + aOtherTargetControl->Width - 1;
   }
   else
      MaxSize = aMaxSize;

   SetCaptureControl(SplitterControl);

   SplitterPos = Point(0, 0);

   if (ResizeVertically)
      OrigPos = aY;
   else
      OrigPos = aX;

   Form->Canvas->Handle = GetDCEx(Form->Handle, 0,
                                  DCX_CACHE|DCX_CLIPSIBLINGS|DCX_LOCKWINDOWUPDATE);

   Form->Canvas->Pen->Color = clWhite;
   Form->Canvas->Pen->Width = 1;
   Form->Canvas->Pen->Mode = pmXor;

   DrawSplitterControl();
}

////////////////////////////////////////////////////////////////////////////////
// OnMouseMove:

void __fastcall TSplitterControl::OnMouseMove(const int &aX, const int &aY)
{
   if (OrigPos != _NO_SIZE_)
   {
      TPoint OldSplitterPos = SplitterPos;

      if (ResizeVertically)
      {
         SplitterPos.y = aY - OrigPos;

         // Check the new coordinates, if required

         if ((TargetControl->Height + SplitterPos.y) < MinSize)
            SplitterPos.y = MinSize - TargetControl->Height;
         else if ((TargetControl->Height + SplitterPos.y) > MaxSize)
            SplitterPos.y = MaxSize - TargetControl->Height;
      }
      else
      {
         SplitterPos.x = aX - OrigPos;

         // Check the new coordinates, if required

         if ((TargetControl->Width + SplitterPos.x) < MinSize)
            SplitterPos.x = MinSize - TargetControl->Width;
         else if ((TargetControl->Width + SplitterPos.x) > MaxSize)
            SplitterPos.x = MaxSize - TargetControl->Width;
      }

      // We draw the new splitter only if there is a need for it

      if ((ResizeVertically && (SplitterPos.y != OldSplitterPos.y)) ||
          (!ResizeVertically && (SplitterPos.x != OldSplitterPos.x)))
      {
         TPoint NewSplitterPos = SplitterPos;

         SplitterPos = OldSplitterPos;

         DrawSplitterControl();

         SplitterPos = NewSplitterPos;

         DrawSplitterControl();
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// OnMouseUp:

void __fastcall TSplitterControl::OnMouseUp(const int &aX, const int &aY)
{
   if (OrigPos != _NO_SIZE_)
   {
      OnMouseMove(aX, aY);

      DrawSplitterControl();

      SetCaptureControl(NULL);

      if (ResizeVertically)
         TargetControl->Height += SplitterPos.y;
      else
         TargetControl->Width += SplitterPos.x;

      OrigPos = _NO_SIZE_;
   }
}

////////////////////////////////////////////////////////////////////////////////
// DrawSplitterControl:

void __fastcall TSplitterControl::DrawSplitterControl()
{
   TPoint Pt = Form->ScreenToClient(SplitterControl->ClientToScreen(SplitterPos));

   Form->Canvas->Rectangle(Pt.x,
                           Pt.y,
                           Pt.x + SplitterControl->Width + 1,
                           Pt.y + SplitterControl->Height + 1);
}

// End of file
////////////////////////////////////////////////////////////////////////////////

