////////////////////////////////////////////////////////////////////////////////
// TTransBitmap component - Oxsoft Heart V5.0
//
// Controls that allows to display a button like with a transparent bitmap, a
// caption, etc. Different ways of displaying both the caption and the picture.
// And... lots of other facilities... :-)
//
// by Alan GARNY
//    gry@physiol.ox.ac.uk
//    http://www.physiol.ox.ac.uk/~gry
//
//  Copyright 1996-97
////////////////////////////////////////////////////////////////////////////////
// Date of Creation: 02/02/97
//
// Modifications: (model: [<initials>, dd/mm/yy] <what has been done>)
// [GRY, 27/02/97] Consider the sub/sup characters when have to break down the
//                 caption. Before it once the caption was broken up, it was
//                 back to normal characters, no matter the previous state of
//                 the characters was (cf. "GetACaptionLine")
// [GRY, 02/03/97] Check that the new font size is coherent with the sub/super
//                 font size difference (cf. "CMFontChanged")
// [GRY, 02/03/97] The [GRY, 27/02/97] fix is not good enough (case of several
//                 sub characters in a row and having to be split up) and has
//                 been rather tricky to set up, therefore an "escape character"
//                 is to be used (by default "\") which allows to manage sub and
//                 super characters without any problem and this doesn't
//                 interfer with "\" if we want to use it ("\\" is to be used in
//                 such a case)
// [GRY, 02/03/97] As for the button look, the clickable panel look allows the
//                 use of up to 3 bitmaps instead of only 1 as the normal panel
//                 look (cf. "Paint" and "GetBitmapWidth")
// [GRY, 08/03/97] Displays the bitmap slightly shifted when the button is down
//                 (when "AllowDown" is set to true) (cf. "GetWorkableRegion")
// [GRY, 15/03/97] Allows the use of a 4th bitmap to be displayed when the mouse
//                 is over the control. Since we allow that, the "FocusWidth"
//                 can be equal to "0"
// [GRY, 15/03/97] From now, the look panel is not an exception, i.e. it behaves
//                 the same way as a clickable panel apart from the mouse events
//                 which are not handled
// [GRY, 15/03/97] Slightly modified the handler for the message "WM_SIZE" which
//                 is "WMSize" by adding a call to "GetBitmapWidth". Indeed,
//                 before if "FAutoSize" was true, the caption's size wasn't
//                 correct and therefore no bitmap was displayed
// [GRY, 28/03/97] Allows to autosize the height only, or the width only or both
//                 (default)
// [GRY, 28/03/97] Removes the spaces that are in the front and in the back of
//                 the caption
// [GRY, 01/04/97] Moves the "HasFocus" property to the "published" section to
//                 make it read-only for the user. This involves the creation
//                 of another property "FHasFocus" that is similar to "HasFocus"
// [GRY, 01/04/97] Updated the "CanvasTextHeight" method to fix a minor bug
//                 about the computation of the height when the sub/super chars
//                 are allowed but there is no caption to display. Before, there
//                 always had a default size (Result = NormalHeight, by default)
// [GRY, 01/04/97] Depending on the type of construction we want to do (dynamic
//                 or not), we set up the properties differently (actually, it
//                 concerns "Width" and "Height"). To do so, the "Initialize"
//                 method has been created. I don't really like that solution,
//                 but I can't see of another one... Any suggestion?? :-)
// [GRY, 02/04/97] Updated to handle the mouse down/move/up events. Now, the
//                 mouse is catched, only if it's really over the clickable part
//                 of the control
// [GRY, 03/04/97] Updated the "Paint" method to check if the mouse is over the
//                 control or not when having to redraw it (useful when a popup
//                 menu has been assigned to the control, though it doesn't
//                 completely solve the problem, cf. below the comment)
// [GRY, 08/04/97] Fixed a small bug in the computation of the height of a part
//                 of the canvas (cf. "CanvasTextHeight"). Indeed, the
//                 computation of the default size that I modified on the 1st of
//                 April 1997 was wrong...
// [GRY, 08/04/97] Added two public methods to simulate the modification of
//                 either the width or the height (cf. "SimulateNewWidth" and
//                 "SimulateNewHeight")
// [GRY, 08/04/97] Allow the control to be automatically refreshed or not
//                 depending on the value of "FAutoRefresh". And in case the
//                 refresh is not allowed, we also have a public method to force
//                 the control to be refreshed (cf. "ForceRefresh")
// [GRY, 09/04/97] Made the "SimulateNewWidth" and "SimulateNewHeight" working.
//                 Indeed, I added them yesterday, but simply didn't work at all
// [GRY, 15/04/97] Allows the dragging (useful when "DragMode" is equal to
//                 "dmManual" to start automatically as soon as the user clicks
//                 on the left mouse button, while pressing the "Ctrl" key
// [GRY, 17/04/97] Small update for the autosize feature. Actually, I forgot one
//                 test in "OptimalSize". Indeed, I considered that if we the
//                 autosize is not only for the width, then it was for the
//                 height only, which is completely wrong, since it can also be
//                 for both...
// [GRY, 18/04/97] Switch a bunch of method from "private" to "protected", so
//                 far I didn't create any class inheriting from TTransBitmap,
//                 so I didn't see the need for it, but now I see it!! :-)
// [GRY, 21/04/97] Added the full transparency property. This way both the
//                 button and the picture can be fully transparent. To get the
//                 full transparency working involved to inherit from
//                 "TGraphicControl" instead of "TCustomControl". This involves
//                 that the control cannot hold other controls anymore... But
//                 I think it's worth the loss!! :-)
// [GRY, 21/04/97] Broke the Paint function up: "DisplayBackground",
//                 "DisplayBitmap", "DisplayCaption" and "DisplayLook", to make
//                 it easier to read and to maintain
// [GRY, 24/04/97] Draws the bitmap differently depending on if the full
//                 transparency is on or not. The main reason is that big
//                 bitmaps that don't need full transparency take really long to
//                 display
// [GRY, 05/05/97] Removed the "csAcceptControls" from the control syle, since
//                 the control cannot hold other controls anymore (since I
//                 implemented the full transparency, cf. 21/04/97)
// [GRY, 12/05/97] Display the bitmap, caption, etc. only if necessary
// [GRY, 13/05/97] Replaced the use of the "" character by the "..." sequence,
//                 indeed using the default font, the "" looks rather strange
// [GRY, 28/05/97] Modified "MouseMove" to make the clickable panel to behave
//                 the same way as the button when it's initially down
// [GRY, 28/05/97] Slighlty modified the capture control process: done only if
//                 the control allows the focus. By doing that, we can allow the
//                 use of keys when over a control that doesn't allow the focus.
//                 However, if the control allows the focus, the use of keys
//                 still doesn't work
// [GRY, 10/06/97] I finally fixed the problem of the "SetCaptureControl"
//                 function which is probably buggy (didn't allow to use the
//                 "TAB" key to navigate through different objects on a form and
//                 the hint didn't show up properly). Anyway, thanks to Neil
//                 Booth (NeilB@Japan.co.jp) who reminding me about the Windows
//                 messages "CM_MOUSEENTER" and "CM_MOUSELEAVE", I finally fixed
//                 that problem!! And god, how neater the code is!!
// [GRY, 12/06/97] Removed the property "ParentColor" which wasn't declared
//                 properly
// [GRY, 12/06/97] Used the propert "Empty" to check whether there is a bitmap
//                 available or not. Before, I simply checked for the size not
//                 to by null... (cf. "Paint") 
// [GRY, 12/06/97] Make the full transparency really full for the display of the
//                 background and of the look (cf. "DisplayBackground" and
//                 "DisplayLook"). Added a new function: "UpdateRectSize"
////////////////////////////////////////////////////////////////////////////////

#include <vcl\vcl.h>
#include <vcl\extctrls.hpp>
#pragma hdrstop

////////////////////////////////////////////////////////////////////////////////

#include "TransBitmap.h"

////////////////////////////////////////////////////////////////////////////////
//
// CONSTRUCTOR & DESTRUCTOR
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Initialize:

void __fastcall TTransBitmap::Initialize(const bool &aDynamic)
{
   FAutoRefresh = true;
   FAutoSizeParams = asBoth;
   FBackgroundColor = clBtnFace;
   FBitmapDisplay = bdMiddleCenter;
   FBitmapNumber = bn1;
   FBevelInner = bvNone;
   FBevelOuter = bvRaised;
   FBevelWidth = 1;
   FCaptionDisplay = cdBitmapTop;
   FCaptionNbLines = 3;
   FEscapeChar = "\\";
   FFocusColor = clHighlight;
   FFocusWidth = 2;
   FLook = loButton;
   FState = stUp;
   FSubChar = "_";
   FSubCharDist = 0;
   FSubFontSizeDiff = 3;
   FSuperChar = "^";
   FSuperCharDist = 0;
   FSuperFontSizeDiff = 3;

   // Control style

   ControlStyle = TControlStyle() << csCaptureMouse << csOpaque <<
                  csDoubleClicks << csReplicatable;

   // Default properties of the control

   Width = Height = aDynamic?0:123;

   FBitmap = new Graphics::TBitmap;   // Internal representation

   CaptionList = new TStringList;

   Canvas->Font = this->Font;

   MonoBitmap = new Graphics::TBitmap;
   NoMonoBitmap = true;
}

////////////////////////////////////////////////////////////////////////////////
// TTransBitmap: constructor

__fastcall TTransBitmap::TTransBitmap(TComponent *aOwner)
   : TGraphicControl(aOwner)
{
   Initialize(false);
}

__fastcall TTransBitmap::TTransBitmap(TComponent *aOwner, const bool &aDynamic)
   : TGraphicControl(aOwner)
{
   Initialize(aDynamic);
}

////////////////////////////////////////////////////////////////////////////////
// ~TTransBitmap: destructor

__fastcall TTransBitmap::~TTransBitmap()
{
   delete CaptionList;

   delete MonoBitmap;

   delete FBitmap;
}

////////////////////////////////////////////////////////////////////////////////
//
// METHOD TO DRAW THE CONTROL
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Paint:

void __fastcall TTransBitmap::Paint()
{
   if (FAutoRefresh)
   {
      // Preliminaries: useful if starting/opening a project... Why? Did I ask
      // you something? No, so... :-)

      CheckSize();

      // Get the region within which we can both draw the bitmap and the caption

      TRect ControlRegion = GetClientRect();
      TRect WorkableRegion = ControlRegion;

      GetWorkableRegion(WorkableRegion);   // Get the real region we can work
                                           // with, which is the client region
                                           // from which we remove the necessary
                                           // space for drawing the look of the
                                           // control

      // Display the background

      if(!FAllowFullTransparency)
         DisplayBackground(WorkableRegion);

      // Display the bitmap

      if (!FBitmap->Empty)
         DisplayBitmap(WorkableRegion);

      // Display the caption

      if (CaptionList->Count)
         DisplayCaption(WorkableRegion);

      // Display the look of the control

      DisplayLook(ControlRegion);
   }
}

////////////////////////////////////////////////////////////////////////////////
//
// METHODS HANDLING DIFFERENT MESSAGES
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// WMSize: manages the resize of the control (both at design and execution time)

void __fastcall TTransBitmap::WMSize(TWMSize &aMsg)
{
   CheckSize();
}

////////////////////////////////////////////////////////////////////////////////
// CMFontChanged: checks out what the new size of the caption is after changing
//                the font, also checks the autosize size if necessary

void __fastcall TTransBitmap::CMFontChanged(TMessage &aMsg)
{
   int PrevFontSize = Canvas->Font->Size;

   Canvas->Font = Font;

   if ((Canvas->Font->Size <= FSubFontSizeDiff) ||
       (Canvas->Font->Size <= FSuperFontSizeDiff))
   {
      Font->Size = PrevFontSize;
      Canvas->Font->Size = PrevFontSize;
   }

   Repaint();
}

////////////////////////////////////////////////////////////////////////////////
// CMTextChanged: checks out what the new size of the caption is after changing
//                it, also checks the autosize size if necessary

void __fastcall TTransBitmap::CMTextChanged(TMessage &aMsg)
{
   // Removes the spaces that are in the front and in the back of the caption

   // In the front first

   while (Caption.SubString(1, 1) == " ")
      Caption = Caption.SubString(2, Caption.Length() - 1);

   // In the back now

   while (Caption.SubString(Caption.Length(), 1) == " ")
      Caption = Caption.SubString(1, Caption.Length() - 1);

   Repaint();
}

////////////////////////////////////////////////////////////////////////////////
// CMMouseEnter: shows that the control got the focus

void __fastcall TTransBitmap::CMMouseEnter(TMessage &aMsg)
{
   FHasFocus = true;

   if (FAllowFocus)
      Repaint();
}

////////////////////////////////////////////////////////////////////////////////
// CMMouseLeave: shows that the control lost the focus

void __fastcall TTransBitmap::CMMouseLeave(TMessage &aMsg)
{
   FHasFocus = false;

   if (FAllowFocus)
      Repaint();
}

////////////////////////////////////////////////////////////////////////////////
//
// METHODS HANDLING A MOUSE CLICK OVER THE CONTROL
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// MouseDown: considers the click of the left mouse button, but only valid if
//            the control is of type "loButton"

void __fastcall TTransBitmap::MouseDown(TMouseButton aButton, TShiftState aShift,
                                        int aX, int aY)
{
   TGraphicControl::MouseDown(aButton, aShift, aX, aY);

   if ((aButton == mbLeft) && (aShift.Contains(ssCtrl)))
      BeginDrag(true);
   else if ((aButton == mbLeft) && (FState != stDisabled) && OverControl(aX, aY))
   {
      MouseCatched = true;   // In order to proceed with the click until it is
                             // released

      // Note: we allow the mouse event to be caught, because it allows to
      //       handle the auto focus for the "normal" panel

      if (FLook != loPanel)
      {
         OrigState = FState;   // Keep the original state in mind, useful if the
                               // control is allowed to stay down

         // If the button is not down, we force it to be that way and then
         // redraw it, thus it looks like it's clicked down
         // This is also valid for the clickable panel

         if (FState != stDown)
         {
            FState = stDown;

            Repaint();
         }
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// MouseMove: if the mouse has been previously catched, we check if the control
//            should be redraw or not, depending on where the mouse is located

void __fastcall TTransBitmap::MouseMove(TShiftState aShift, int aX, int aY)
{
   TGraphicControl::MouseMove(aShift, aX, aY);

   // Check if the control has the focus or not

   bool NeedRepaint = false;

   if (MouseCatched)
   {
      if (((FLook == loButton) || (FLook == loClickablePanel)) &&
          (!FAllowDown || (OrigState != stDown)))
      {
         // Note that we consider the mouse move only if that's a clickable
         // panel or a button which is not allowed to be down or that its
         // original state wasn't down
         // This allows the clickable panel or button which is initially down to
         // stay down

         // We check out what the state of the button should be, depending on
         // the mouse's position

         TTransBitmapState NewState = (OverControl(aX, aY))?stDown:stUp;

         // If the state has changed, we set it to its new state and redraw the
         // control

         if (NewState != FState)
         {
            FState = NewState;

            NeedRepaint = true;
         }
      }
   }

   if (NeedRepaint)
      Repaint();
}

////////////////////////////////////////////////////////////////////////////////
// MouseUp: ends the clicking of the button if it has been started, the clicking
//          involves the execution of the "OnClick" event if the mouse button is
//          released over the control

void __fastcall TTransBitmap::MouseUp(TMouseButton aButton, TShiftState aShift,
                                      int aX, int aY)
{
   TGraphicControl::MouseUp(aButton, aShift, aX, aY);

   if (MouseCatched)
   {
      MouseCatched = false;   // The process is terminated

      // Determine the new state of the control

      if (FAllowDown && (FLook != loPanel))
      {
         // The button/clickable panel is allowed to stay down, we therefore
         // need to check what was the original state ("stUp" or "stDown") and
         // depending on it set the new one to the opposite (resp. "stDown" or
         // "stUp")

         if (OverControl(aX, aY))
            FState = (OrigState == stDown)?stUp:stDown;
      }
      else
         // Typical button or clickable panel, we therefore come back to a
         // typical state: "stUp"

         FState = stUp;

      // If we released the mouse button over the control then we call the
      // "OnClick" event

      if (OverControl(aX, aY))
      {
         if (OnClick || (FLook != loPanel))
         {
            Repaint();

            Click();   // Will do nothing if there is no click handler
         }
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
//
// METHODS TO MODIFY THE DIFFERENT PUBLISHED PROPERTIES
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// SetAllowDown: allows or not the control to stay down when being of typ
//               "loButton"

void __fastcall TTransBitmap::SetAllowDown(const bool aValue)
{
   if (aValue != FAllowDown)
   {
      FAllowDown = aValue;

      if (!FAllowDown && (FState == stDown))
      {
         // If "FAllowDown" is false, we can't afford to have "FState" set to
         // "stDown", we therefore set it to "stUp"

         FState = stUp;

         Repaint();
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetAllowFocus: allows or not the control to show if the mouse is over it or
//                not

void __fastcall TTransBitmap::SetAllowFocus(const Boolean aValue)
{
   if (aValue != FAllowFocus)
   {
      FAllowFocus = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetAllowFullTransparency: allows or not the control to have a fully
//                           transparent background 

void __fastcall TTransBitmap::SetAllowFullTransparency(const Boolean aValue)
{
   if (aValue != FAllowFullTransparency)
   {
      FAllowFullTransparency = aValue;

      if (FAllowFullTransparency)
         ControlStyle >> csOpaque;
      else
         ControlStyle << csOpaque;

      Repaint();  
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetAllowDown: allows or not the use of sub/super characters

void __fastcall TTransBitmap::SetAllowSubAndSuperChars(const bool aValue)
{
   if (aValue != FAllowSubAndSuperChars)
   {
      FAllowSubAndSuperChars = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetAutoRefresh: allows or not the control to be refreshed automatically

void __fastcall TTransBitmap::SetAutoRefresh(const Boolean aValue)
{
   if (aValue != FAutoRefresh)
   {
      FAutoRefresh = aValue;

      if (FAutoRefresh)
         Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetAutoSize: enables or not the "FAutoSize" property, if enabled then forces
//              the control to resize and draw itself

void __fastcall TTransBitmap::SetAutoSize(const bool aValue)
{
   if (aValue != FAutoSize)
   {
      FAutoSize = aValue;

      if (FAutoSize)   // No need to repaint if it's not necessary...
         Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetAutoSizeParams: specifies what has to be autosized in case this is allowed

void __fastcall TTransBitmap::SetAutoSizeParams(const TTransBitmapAutoSize aValue)
{
   if (aValue != FAutoSizeParams)
   {
      FAutoSizeParams = aValue;

      if (FAutoSize)   // No need to repaint if it's not necessary...
         Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBackgroundColor: modifies the color for the background. Only used when the
//                     control is of typed "loPanel" or "loClickablePanel"

void __fastcall TTransBitmap::SetBackgroundColor(const TColor aValue)
{
   if (aValue != FBackgroundColor)
   {
      FBackgroundColor = aValue;

      if (FLook != loButton)
         Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBevelInner: changes the type of the bevel inner, therefore we need to
//                redraw the component, but first we need to check the new size
//                of the control and check for realignment

void __fastcall TTransBitmap::SetBevelInner(const TPanelBevel aValue)
{
   if (aValue != FBevelInner)
   {
      FBevelInner = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBevelOuter: changes the type of the bevel outer, therefore we need to
//                redraw the component, but first we need to check the new size
//                of the control and check for realignment

void __fastcall TTransBitmap::SetBevelOuter(const TPanelBevel aValue)
{
   if (aValue != FBevelOuter)
   {
      FBevelOuter = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBevelWidth: changes the type of the bevel width, therefore we need to
//                redraw the component, but first we need to check the new size
//                of the control and check for realignment

void __fastcall TTransBitmap::SetBevelWidth(const int aValue)
{
   if ((aValue != FBevelWidth) && (aValue > 0))
   {
      FBevelWidth = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBitmap: loads a bitmap, therefore we need to check for the size and then
//            redraw the control

void __fastcall TTransBitmap::SetBitmap(Graphics::TBitmap *aValue)
{
   FBitmap->Assign(aValue);

   NoMonoBitmap = true;

   Repaint();
}

////////////////////////////////////////////////////////////////////////////////
// SetBitmapDisplay: changes the way to display the bitmap, therefore we need
//                   to redraw it

void __fastcall TTransBitmap::SetBitmapDisplay(const TTransBitmapBitmapDisplay aValue)
{
   if (aValue != FBitmapDisplay)
   {
      FBitmapDisplay = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBitmapNumber: changes the number of bitmaps hold by "FBitmap", therefore
//                  we need to know what the new width is for the bitmap, and
//                  since the width can change, we need to check the new size
//                  for the control, and finally to redraw it

void __fastcall TTransBitmap::SetBitmapNumber(const TTransBitmapBitmapNumber aValue)
{
   if (aValue != FBitmapNumber)
   {
      FBitmapNumber = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetBorderWidth: changes the size of the border, consequently we must check
//                 the new size of the control, the realignment and finally
//                 redraw the control

void __fastcall TTransBitmap::SetBorderWidth(const int aValue)
{
   if ((aValue != FBorderWidth) && (aValue >= 0))
   {
      FBorderWidth = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetCaptionDisplay: changes the way to display the caption, and since it may
//                    be displayed in function of the bitmap, we therefore need
//                    to check the size and redraw the control

void __fastcall TTransBitmap::SetCaptionDisplay(const TTransBitmapCaptionDisplay aValue)
{
   if (aValue != FCaptionDisplay)
   {
      FCaptionDisplay = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetCaptionNbLines: allows to decide how many lines can be used to display the
//                    caption, we therefore need to check the new size of the
//                    caption and consequently of the control (necessary when
//                    "FAutoSize" is TRUE), and finally once this is done, we
//                    need to redraw the control itself

void __fastcall TTransBitmap::SetCaptionNbLines(const int aValue)
{
   if ((aValue != FCaptionNbLines) && (aValue >= 1))
   {
      FCaptionNbLines = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetEscapeChar: changes the escape character

void __fastcall TTransBitmap::SetEscapeChar(const String aValue)
{
   if (aValue != FEscapeChar)
   {
      FEscapeChar = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetFocusColor: changes the color to be used when the control gets the focus

void __fastcall TTransBitmap::SetFocusColor(const TColor aValue)
{
   if (aValue != FFocusColor)
   {
      FFocusColor = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetFocusWidth: changes the width of the focus frame

void __fastcall TTransBitmap::SetFocusWidth(const int aValue)
{
   if ((aValue != FFocusWidth) && (aValue >= 0))
   {
      FFocusWidth = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetLook: switches from a panel look/behaviour to a button look/behaviour (and
//          vice-versa), we therefore need to check the new control's size,
//          before redrawing it

void __fastcall TTransBitmap::SetLook(const TTransBitmapLook aValue)
{
   if (aValue != FLook)
   {
      FLook = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetState: used to specify the current state of the control when being of type
//           "loButton", nevertheless we allow the modification even if of type
//           "loPanel". If "FAllowDown" is FALSE, we don't allow the state to be
//           set to "stDown". The value of the "Enabled" property depends on
//           both its value and the new value of the state. Finally, we redraw
//           the control accordingly to the new state

void __fastcall TTransBitmap::SetState(const TTransBitmapState aValue)
{
   if (aValue != FState)
   {
      if ((aValue == stDown) && !FAllowDown)
         FState = (FState == stUp)?stDisabled:stUp;
      else
         FState = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetStretch: forces the bitmap to be stretched or not, this just involves to
//             repaint the control

void __fastcall TTransBitmap::SetStretch(const Boolean aValue)
{
   if (aValue != FStretch)
   {
      FStretch = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetSubChar: specifies which character has to be used as a tag for writing sub
//             characters

void __fastcall TTransBitmap::SetSubChar(const String aValue)
{
   if ((aValue != FSubChar) && (aValue.Length() == 1))
   {
      FSubChar = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetSubDist: specifies the distance to which the sub characters have to be
//             written

// Note: it cannot be located above the normal text

void __fastcall TTransBitmap::SetSubCharDist(const int aValue)
{
   if ((aValue != FSubCharDist) && (aValue >= 0))
   {
      FSubCharDist = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetSubFontSizeDiff: specifies how small the font to be used for the sub
//                     characters has to be in comparison with the normal one

void __fastcall TTransBitmap::SetSubFontSizeDiff(const int aValue)
{
   if ((aValue != FSubFontSizeDiff) && (aValue >= 1) && 
       (aValue < Canvas->Font->Size))
   {
      FSubFontSizeDiff = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetSuperChar: specifies which character has to be used as a tag for writing
//               super characters

void __fastcall TTransBitmap::SetSuperChar(const String aValue)
{
   if ((aValue != FSuperChar) && (aValue.Length() == 1))
   {
      FSuperChar = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetSuperDist: specifies the distance to which the super characters have to be
//               written

// Note: it cannot be located below the normal text

void __fastcall TTransBitmap::SetSuperCharDist(const int aValue)
{
   if ((aValue != FSuperCharDist) && (aValue >= 0))
   {
      FSuperCharDist = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
// SetSuperFontSizeDiff: specifies how small the font to be used for the super
//                       characters has to be in comparison with the normal one

void __fastcall TTransBitmap::SetSuperFontSizeDiff(const int aValue)
{
   if ((aValue != FSuperFontSizeDiff) && (aValue >= 1) &&
       (aValue < Canvas->Font->Size))
   {
      FSuperFontSizeDiff = aValue;

      Repaint();
   }
}

////////////////////////////////////////////////////////////////////////////////
//
// METHODS USED FOR INTERNAL PURPOSE
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// DisplayBackground: displays the background of the control if required (i.e.
//                    the full transparency is not allowed)

void __fastcall TTransBitmap::DisplayBackground(const TRect &aRegion)
{
   // The background color depends on the type of the control
   // If the control is of type "loButton", we consider the state of the
   // control, and the fact if it's currently catching a mouse click or
   // not

   if (FLook != loButton)
      Canvas->Brush->Color = FBackgroundColor;
   else if (((FState == stDown) && !MouseCatched) ||
            ((OrigState == stDown) && MouseCatched))
      Canvas->Brush->Color = clBtnHighlight;
   else
      Canvas->Brush->Color = clBtnFace;

   TRect RealRegion = aRegion;

   if (FAllowFullTransparency && !FHasFocus)
      UpdateRectSize(RealRegion, -FFocusWidth);   // This keeps the background
                                                  // as it is

   Canvas->FillRect(RealRegion);   // Draw the background
}

////////////////////////////////////////////////////////////////////////////////
// DisplayBitmap: displays the bitmap according to the user's specifications

void __fastcall TTransBitmap::DisplayBitmap(const TRect &aRegion)
{
   FAllowFullTransparency?
      DisplayBitmapWithFullTransparency(aRegion):
      DisplayBitmapWithoutFullTransparency(aRegion);
}

////////////////////////////////////////////////////////////////////////////////
// DisplayBitmapWithFullTransparency: displays the bitmap using a full
//                                    transparency

void __fastcall TTransBitmap::DisplayBitmapWithFullTransparency(const TRect &aRegion)
{
   try
   {
      Graphics::TBitmap *SelBmp;   // Bitmap that corresponds to the one to
                                   // display depending on the state of the
                                   // control
      Graphics::TBitmap *MonoBmp;   // Monochrome version of SelBmp
                                    // Note: "MonoBmp" is to be changed during
                                    //       the procedure: it will be inverted.
                                    //       Useful to get the full transparency
      Graphics::TBitmap *ResBmp;   // Resulting bitmap that has to be displayed
      TRect DestBmpRegion, SrcBmpRegion;   // Used for getting the right bitmap
                                           // according to the number of them,
                                           // the state and the type of the
                                           // control
      TPoint BmpPt;   // Position of the bitmap

      SelBmp  = new Graphics::TBitmap;
      MonoBmp = new Graphics::TBitmap;
      ResBmp  = new Graphics::TBitmap;

      // We select the right bitmap to display depending on the number of
      // bitmaps and the current state of the control

      // Get the region of where the bitmap is to be displayed

      if (FStretch)
      {
         // The region where to display the stretched picture depends on where
         // the caption is to be displayed (with the origin at (0, 0))

         DestBmpRegion = aRegion;

         DestBmpRegion.Right  -= DestBmpRegion.Left;
         DestBmpRegion.Bottom -= DestBmpRegion.Top;
         DestBmpRegion.Left    = 0;
         DestBmpRegion.Top     = 0;

         // Slight update in some cases which depend on the position of the
         // caption

         switch (FCaptionDisplay)
         {
            case cdBitmapLeft:
            case cdBitmapRight:
               DestBmpRegion.Right -= CaptionWidth; break;
            case cdBitmapTop:
            case cdBitmapBottom:
               DestBmpRegion.Bottom -= CaptionHeight; break;
         }
      }
      else
         DestBmpRegion = Rect(0, 0, BitmapWidth, FBitmap->Height);

      // Determine the size of the region where the bitmap is to be displayed

      int DestWidth  = DestBmpRegion.Right - DestBmpRegion.Left;
      int DestHeight = DestBmpRegion.Bottom - DestBmpRegion.Top;

      // Initialise the size of the different bitmaps

      SelBmp->Width  = MonoBmp->Width  = ResBmp->Width  = DestWidth;
      SelBmp->Height = MonoBmp->Height = ResBmp->Height = DestHeight;

      // Get the region of where the bitmap to display is located (within
      // "FBitmap")

      GetSubBitmap(SelBmp, DestBmpRegion, SrcBmpRegion);

      // Generate the monochrome version of "FBitmap" (if necessary) and get
      // the sub part we need

      ComputeMonoBitmap();
      MonoBmp->Canvas->CopyRect(DestBmpRegion, MonoBitmap->Canvas, SrcBmpRegion);

      // Get the position at which the bitmap will have to be displayed at the
      // end

      if (FStretch)
      {
         // Final position of the bitmap: depends on where the caption is to be
         // displayed
         BmpPt.x = aRegion.Left;   // Default values for the position of the
         BmpPt.y = aRegion.Top;    // stretched bitmap

         // Slight update in certain cases

         switch (FCaptionDisplay)
         {
            case cdBitmapRight: BmpPt.x += CaptionWidth; break;
            case cdBitmapBottom: BmpPt.y += CaptionHeight; break;
         }
      }
      else
         GetBitmapPos(aRegion, BmpPt);

      // Copy into "ResBmp" what is under the control, but first we have to find
      // out the location and size of the final bitmap

      SrcBmpRegion = DestBmpRegion;   // Note: useful only if the bitmap is
                                      //       stretched, since "SrcBmpRegion"
                                      //       equals "DestBmpRegion" normally

      SrcBmpRegion.Top    += BmpPt.y;   // A slight correction to the region is
      SrcBmpRegion.Left   += BmpPt.x;   // necessary to take into account the
      SrcBmpRegion.Right  += BmpPt.x;   // final position of the control
      SrcBmpRegion.Bottom += BmpPt.y;

      ResBmp->Canvas->CopyRect(DestBmpRegion, Canvas, SrcBmpRegion);

      // Some bits' operations to get the full transparency:
      //  - Get the mask of the bitmap on top of the background
      //  - Get the inverse of the monochrome bitmap,
      //  - Get the bitmap without all the surrounding (this using the inverse
      //    of the monochrome bitmap),
      //  - Copy the result of the the third operation on top of the result of
      //    the first operation

      BitBlt(ResBmp->Canvas->Handle, 0, 0, DestWidth, DestHeight, MonoBmp->Canvas->Handle, 0, 0, SRCAND);
      BitBlt(MonoBmp->Canvas->Handle, 0, 0, DestWidth, DestHeight, MonoBmp->Canvas->Handle, 0, 0, NOTSRCCOPY);
      BitBlt(SelBmp->Canvas->Handle, 0, 0, DestWidth, DestHeight, MonoBmp->Canvas->Handle, 0, 0, SRCAND);
      BitBlt(ResBmp->Canvas->Handle, 0, 0, DestWidth, DestHeight, SelBmp->Canvas->Handle, 0, 0, SRCPAINT);

      // At last, we can display the bitmap!! :-)

      Canvas->Draw(BmpPt.x, BmpPt.y, ResBmp);

      delete ResBmp;
      delete MonoBmp;
      delete SelBmp;
   } catch(...) {}
}

////////////////////////////////////////////////////////////////////////////////
// DisplayBitmapWithoutFullTransparency: displays the bitmap without using a
//                                       full transparency

void __fastcall TTransBitmap::DisplayBitmapWithoutFullTransparency(const TRect &aRegion)
{
   try
   {
      Graphics::TBitmap *TmpBmp;   // Corresponds to a subdivision of "FBitmap"
                                   // (in case there are more than 1 sub-pics)

      TmpBmp = new Graphics::TBitmap;

      // Initialise the size of the temporary bitmap
      // Note: the width may be too big, but it doesn't really matter. Actually,
      //       it's better that way that the other way round...

      TmpBmp->Width  = BitmapWidth;
      TmpBmp->Height = FBitmap->Height;

      // We select the right bitmap to display depending on the number of
      // bitmaps and the current state of the control

      TRect TmpBmpRegion;

      GetSubBitmap(TmpBmp, Rect(0, 0, TmpBmp->Width, TmpBmp->Height), TmpBmpRegion);

      // Make the bitmap transparent if a transparent color exists within the
      // bitmap

      TmpBmp->Canvas->Brush->Color = Canvas->Brush->Color;

      TmpBmp->Canvas->BrushCopy(Rect(0, 0, TmpBmp->Width, TmpBmp->Height), TmpBmp,
                                Rect(0, 0, TmpBmp->Width, TmpBmp->Height), TmpBmp->TransparentColor);

      // Display the bitmap itself

      // Display the bitmap, 2 cases: 1) normal, 2) stretch

      if (FStretch)
      {
         // The region to use for stretching depends on if the caption is
         // displayed relatively to the bitmap or not

         TRect StretchRegion = aRegion;

         switch (FCaptionDisplay)
         {
            case cdBitmapLeft:
               StretchRegion.Right -= CaptionWidth; break;
            case cdBitmapRight:
               StretchRegion.Left += CaptionWidth; break;
            case cdBitmapTop:
               StretchRegion.Bottom -= CaptionHeight; break;
            case cdBitmapBottom:
               StretchRegion.Top += CaptionHeight; break;
         }

         Canvas->StretchDraw(StretchRegion, TmpBmp);
      }
      else
      {
         // We display the bitmap using its normal size
         // The location depends on both the property "FBitmapDisplay" and
         // "FCaptionDisplay"

         TPoint BmpPt;

         GetBitmapPos(aRegion, BmpPt);

         Canvas->Draw(BmpPt.x, BmpPt.y, TmpBmp);
      }

      delete TmpBmp;
   } catch(...) {}
}

////////////////////////////////////////////////////////////////////////////////
// GetSubBitmap: gets the bitmap within "FBitmap" that corresponds to the
//               current state

void __fastcall TTransBitmap::GetSubBitmap(Graphics::TBitmap *aDestBmp,
                                           const TRect &aDestRegion,
                                           TRect &aSrcRegion)
{
   // Get the region of where the bitmap to display is located (within
   // "FBitmap")

   aSrcRegion = Rect(0, 0, BitmapWidth, FBitmap->Height);   // Default 

   switch (FState)
   {
      case stDisabled:
         if (FBitmapNumber >= bn2)   // A disabled bitmap is available
            aSrcRegion = Rect(BitmapWidth, 0, 2 * BitmapWidth, FBitmap->Height);

         break;
      case stDown:
         if (FBitmapNumber >= bn3)   // A down bitmap is available
            aSrcRegion = Rect(2 * BitmapWidth, 0, 3 * BitmapWidth, FBitmap->Height);

         break;
   }

   if (FAllowFocus && FHasFocus && (FBitmapNumber == bn4))
      // The control has the focus and it is allowed to get it, moreover
      // there is a picture available, therefore...

      aSrcRegion = Rect(3 * BitmapWidth, 0, 4 * BitmapWidth, FBitmap->Height);

   // Get the right bitmap, both the "normal" one and the monochrome

   aDestBmp->Canvas->CopyRect(aDestRegion, FBitmap->Canvas, aSrcRegion);
}

////////////////////////////////////////////////////////////////////////////////
// DisplayCaption: displays the caption according to the user's specifications

void __fastcall TTransBitmap::DisplayCaption(const TRect &aRegion)
{
   TPoint TxtPt;   // Position of the first caption line
   int CrtCaptionLine;   // Used to display the different lines which compound
                         // the caption
   int CaptionLineTop;   // Used to display the different lines which compound
                         // the caption

   // Display the caption with a transparent background and in several times if
   // necessary (i.e. the caption is too long) and required ("FCaptionNbLines"
   // property)

   Canvas->Brush->Style = bsClear;   // Allows to have a transparent
                                     // background for the caption

   GetCaptionPos(aRegion, TxtPt);   // Get the position of the first line of
                                    // the caption

   CaptionLineTop = TxtPt.y;

   // Displays the line(s) which compound(s) the caption

   for (CrtCaptionLine = 0; CrtCaptionLine < CaptionList->Count; CrtCaptionLine++)
   {
      CanvasTextOut(TxtPt.x + (CaptionWidth - CanvasTextWidth(CaptionList->Strings[CrtCaptionLine])) / 2, CaptionLineTop,
                    CaptionList->Strings[CrtCaptionLine]);

      CaptionLineTop += CanvasTextHeight(CaptionList->Strings[CrtCaptionLine]);
   }
}

////////////////////////////////////////////////////////////////////////////////
// DisplayLook: displays the look of the control

void __fastcall TTransBitmap::DisplayLook(TRect &aRegion)
{
   TColor TopColor, BottomColor;   // Colors for the 3D effect

   if (FAllowFocus && FFocusWidth)
   {
      if (FHasFocus)
         Frame3D(Canvas, aRegion, FFocusColor, FFocusColor, FFocusWidth);
      else if (!FAllowFullTransparency)
         Frame3D(Canvas, aRegion, clBtnFace, clBtnFace, FFocusWidth);
      else
         UpdateRectSize(aRegion, -FFocusWidth);   // This keeps the background
                                                  // as it is 
   }

   if (FLook == loButton)
      // The button consists of two frames with different colors in function of
      // the state

      switch (FState)
      {
         case stUp:
         case stDisabled:
            Frame3D(Canvas, aRegion, clBtnHighlight, clBlack, 1);

            Canvas->Pen->Color = clBtnShadow;

            Canvas->MoveTo(aRegion.Left, aRegion.Bottom - 1);
            Canvas->LineTo(aRegion.Right - 1, aRegion.Bottom - 1);
            Canvas->LineTo(aRegion.Right - 1, aRegion.Top - 1);

            break;
         case stDown:
            Frame3D(Canvas, aRegion, clBlack, clBtnHighlight, 1);

            Canvas->Pen->Color = clBtnShadow;

            Canvas->MoveTo(aRegion.Left, aRegion.Bottom - 1);
            Canvas->LineTo(aRegion.Left, aRegion.Top);
            Canvas->LineTo(aRegion.Right, aRegion.Top);

            break;
      }
   else
   {
      // The panel is drawn in 3 steps: 1) outside, 2) border and 3) inside

      if (FBevelOuter != bvNone)
      {
         if ((FLook == loClickablePanel) && (FState == stDown))
            AdjustColors(FBevelOuter, BottomColor, TopColor);
         else
            AdjustColors(FBevelOuter, TopColor, BottomColor);

         Frame3D(Canvas, aRegion, TopColor, BottomColor, BevelWidth);
      }

      if (!FAllowFullTransparency)
         Frame3D(Canvas, aRegion, FBackgroundColor, FBackgroundColor, BorderWidth);
      else
         UpdateRectSize(aRegion, -BorderWidth);   // This allows to keep the
                                                  // background as it is

      if (FBevelInner != bvNone)
      {
         if ((FLook == loClickablePanel) && (FState == stDown))
            AdjustColors(FBevelInner, BottomColor, TopColor);
         else
            AdjustColors(FBevelInner, TopColor, BottomColor);

         Frame3D(Canvas, aRegion, TopColor, BottomColor, BevelWidth);
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// ComputeMonoBitmap: computes the monochrome version of the bitmap

void __fastcall TTransBitmap::ComputeMonoBitmap()
{
   if (NoMonoBitmap)
   {
      MonoBitmap->Width = FBitmap->Width;
      MonoBitmap->Height = FBitmap->Height;

      for (int i = 0; i < FBitmap->Width; i++)
         for (int j = 0; j < FBitmap->Height; j++)
            if (FBitmap->Canvas->Pixels[i][j] == (FBitmap->TransparentColor&~0x02000000))
               MonoBitmap->Canvas->Pixels[i][j] = clWhite;
            else
               MonoBitmap->Canvas->Pixels[i][j] = clBlack;

      NoMonoBitmap = false;
   }
}

////////////////////////////////////////////////////////////////////////////////
// OverControl: returns true if the coordinates are within the clickable part of
//              the control

bool __fastcall TTransBitmap::OverControl(const int &aX, const int &aY)
{
   TRect Region = GetClientRect();

   // Extract the clickable part of the control

   if (FAllowFocus)
      UpdateRectSize(Region, -FFocusWidth);

   return ((aX >= Region.Left) && (aX <= Region.Right) &&
           (aY >= Region.Top) && (aY <= Region.Bottom));
}

bool __fastcall TTransBitmap::OverControl(const TPoint &aPoint)
{
   return OverControl(aPoint.x, aPoint.y);
}

////////////////////////////////////////////////////////////////////////////////
// AdjustColors: finds out what colors should be used to display the 3D
//               representation of the panel look like. This depends on the
//               bevel's type ("bvLowered" or "bvRaised")

void __fastcall TTransBitmap::AdjustColors(const TPanelBevel &aBevel,
                                           TColor &aTopColor,
                                           TColor &aBottomColor)
{
   if (aBevel == bvLowered)
   {
      aTopColor = clBtnShadow;
      aBottomColor = clBtnHighlight;
   }
   else
   {
      aTopColor = clBtnHighlight;
      aBottomColor = clBtnShadow;
   }
}

////////////////////////////////////////////////////////////////////////////////
// GetWorkableRegion: finds out which region within the control can be used to
//                    display the bitmap/caption. To get it, we need to remove
//                    all the space necessary to represent the control in a
//                    panel or button look like

void __fastcall TTransBitmap::GetWorkableRegion(TRect &aRegion)
{
   int Delta = 0;   // Nb of pixels to remove from left, right, top and bottom
                    // to get the workable region

   if (FAllowFocus)
      Delta += FFocusWidth;

   if (FLook == loButton)
      // Button look

      Delta += 2;
   else
   {
      // Panel look

      if (FBevelOuter != bvNone)
         Delta += FBevelWidth;

      Delta += FBorderWidth;

      if (FBevelInner != bvNone)
         Delta += FBevelWidth;
   }

   // Generate the workable region: consists of removing the useless part from
   // the original region

   UpdateRectSize(aRegion, -Delta);

   // Slight update if that's a button: shift the display of the bitmap and the
   // caption

   if (FLook == loButton)
   {
      if (FState == stDown)
      {
         aRegion.Bottom++;
         aRegion.Right++;
      }
      else
      {
         aRegion.Top--;
         aRegion.Left--;
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// CrtWidth: returns the length of part of the caption

int __fastcall TTransBitmap::CrtWidth(const String &aCaptionString,
                                      const int &aLimit,
                                      const int &aCaptionLength,
                                      const bool &aLastCaptionLine)
{
   if ((aLimit == aCaptionLength) || !aLastCaptionLine)
      return CanvasTextWidth(aCaptionString.SubString(1, aLimit));
   else
      return CanvasTextWidth(aCaptionString.SubString(1, aLimit) + "...");
}

////////////////////////////////////////////////////////////////////////////////
// GetACaptionLine: according to the width of the workable region and the
//                  caption to display, we find out what is the longest part of
//                  the caption which can be displayed. If we have to find out
//                  the last part to display, we append the "..." sequence of
//                  characters at it

String __fastcall TTransBitmap::GetACaptionLine(String &aCaptionString,
                                                const int &aRegionWidth,
                                                const bool &aLastCaptionLine)
{
   int I, J;
   int CaptionLength = aCaptionString.Length();
   String Result = "";

   if (CanvasTextWidth(aCaptionString) <= aRegionWidth)
   {
      // Easy case: the caption string fits within the region

      Result = aCaptionString;

      aCaptionString = "";
   }
   else
   {
      // Unlucky: the caption string doesn't fit, so we have to split it up by
      // looking for the " " character

      I = aCaptionString.Pos(" ");

      if (!I || (I && (CrtWidth(aCaptionString, I - 1, CaptionLength, aLastCaptionLine) > aRegionWidth)))
      {
         // No " " character or there is one but too far, so we need to cut the
         // string whenever it is necessary and maybe append "..."

         I = 0;

         do
            I++;
         while ((I <= CaptionLength) && (CrtWidth(aCaptionString, I, CaptionLength, aLastCaptionLine) <= aRegionWidth));

         I--;   // The last character is too much or we have reached the end
      }
      else
      {
         // There is at least one space within the caption, but we have to check
         // out if that the optimal one or not

         J = 0;

         while ((I != J) && (CrtWidth(aCaptionString, I - 1, CaptionLength, aLastCaptionLine) <= aRegionWidth))
         {
            J = I;

            I = aCaptionString.SubString(I + 1, aCaptionString.Length() - I).Pos(" ") + I;
         } while ((I != J) && (CrtWidth(aCaptionString, I - 1, CaptionLength, aLastCaptionLine) <= aRegionWidth));

         // Either there is no " " character left or the next one is too far
         // away, we therefore have to use the previous " " character

         I = J - 1;   // Just before the " " character
      }

      if ((I == CaptionLength) || !aLastCaptionLine)
      {
         Result = aCaptionString.SubString(1, I);

         // And the reminder
         // Note: we have to had the sub or super marker, if necessary, i.e. if
         //       the control allows sub/super characters and if the caption
         //       line we just computed ends with a sub/super character

         if (aCaptionString.SubString(I + 1, 1) == " ")
            // We skip the first character, since that's a blank one
            // Note: we consider the user clever enough for not writing several
            //       blank characters in a row

            aCaptionString = aCaptionString.SubString(I + 2, aCaptionString.Length() - I);
         else
            aCaptionString = aCaptionString.SubString(I + 1, aCaptionString.Length() - I);

         // We add the sub or super marker in front of the remaining caption, in
         // case this is necessary, i.e. if the control allows sub/super chars
         // and if the caption line we just computed ends with a sub/super char

         if (FAllowSubAndSuperChars)
         {
            // Analyse the caption line we just computed (i.e. Result) to check
            // out if it ends with a sub, super or normal character

            int EndsWith = 0;   // 0 = normal, 1 = sub and 2 = super
            int ResLength = Result.Length();
            String CharToTest;
            bool CanConsiderLastChar;

            for (I = 1; I < ResLength; I++)
            {
               CharToTest = Result.SubString(I, 1);

               if (CharToTest == FEscapeChar)
                  // That's the escape character, therefore we just skip it
                  // Note: the following one is also to be skipped, but it's
                  //       done automatically done at the end of the loop

                  I ++;
               else if (CharToTest == FSubChar)   // Switch to sub/normal mode
                  EndsWith = (EndsWith == 1)?0:1;
               else if (CharToTest == FSuperChar)   // Switch to super/normal
                                                    // mode
                  EndsWith = (EndsWith == 2)?0:2;

               CanConsiderLastChar = (CharToTest != FEscapeChar);
            }

            // Consider the last character of the caption line

            if (CanConsiderLastChar)
            {
               CharToTest = Result.SubString(ResLength, 1);

               if (CharToTest == FEscapeChar)
               {
                  // That's the escape character, therefore we have to skip it
                  // for the current caption line, but keep it in mind for the
                  // remaining

                  Result = Result.SubString(1, ResLength - 1);
                  aCaptionString = FEscapeChar + aCaptionString;
               }
               else if (CharToTest == FSubChar)   // Switch to sub/normal mode
                  EndsWith = (EndsWith == 1)?0:1;
               else if (CharToTest == FSuperChar)   // Switch to super/normal
                                                    // mode
                  EndsWith = (EndsWith == 2)?0:2;
            }

            // At this point, we know in which mode the caption line ends and
            // how it ends, consequently, we can know what to do with the
            // remaining (i.e. put the sub or super mark in front of it or not)

            switch (EndsWith)
            {
               case 1:   // Sub
                  aCaptionString = FSubChar + aCaptionString; break;
               case 2:   // Super
                  aCaptionString = FSuperChar + aCaptionString; break;
            }
         }
      }
      else
         // That's the last caption line we have to compute, nevertheless there
         // are till some characters that should be displayed, consequently we
         // add the "..." character

         Result = aCaptionString.SubString(1, I) + "...";
   }

   return Result;
}

////////////////////////////////////////////////////////////////////////////////
// ComputeCaptionSize: computes the necessary space to display the caption

void __fastcall TTransBitmap::ComputeCaptionSize(const bool &aHeightLimit,
                                                 const bool &aSimulation)
{
   TRect Region;   // Holds the workable region from which we can determine the
                   // required space to display the caption
   int RegionWidth, RegionHeight;   // Width and Height of the workable region
   int NbToProcess;   // Number of lines to retrieve (in relation with
                      // "FCaptionNbLines"
   int CrtLine;   // Current caption line we work with to determine the caption
                  // width
   String TmpCaption;   // Temporary caption we use to find out the size
   int CrtCaptionWidth;   // Width of the current line of the caption
   bool LastOne;   // To know if we want to determine the last caption line or
                   // not
   int DefFontSize = Canvas->Font->Size;   // Size of a the font for a norm char

   // Compute the height for a normal, sub and super character

   NormalHeight = Canvas->TextHeight("X");

   Canvas->Font->Size = DefFontSize - FSubFontSizeDiff;
   SubHeight = Canvas->TextHeight("X");

   Canvas->Font->Size = DefFontSize - FSuperFontSizeDiff;
   SuperHeight = Canvas->TextHeight("X");

   Canvas->Font->Size = DefFontSize;   // Back to the default size

   // Default width and height of the caption

   CaptionWidth = CanvasTextWidth(Caption);
   CaptionHeight = CanvasTextHeight(Caption);

   // Remove all the caption lines from the list to be cleaned

   CaptionList->Clear();

   // We are only going to find out the size if there is something to find
   // out (i.e. there is a caption)

   if (Caption != "")
   {
      if (FAutoSize && (FAutoSizeParams != asHeightOnly))
         // Since "FAutoSize" is set to TRUE, we therefore have nothing to
         // find out, because the size will be the one of the raw caption
         // The only thing to do is to add the caption to the list and that
         // is

         CaptionList->Add(Caption);
      else
      {
         // The control is not autosized, we therefore have to find out the
         // required size for displaying the caption

         // Get the size of the area within which we can display the caption

         Region = GetClientRect();

         // Update the region in case we want to make a simulation

         if (aSimulation)
         {
            Region.Right  += SimWidth - Width;
            Region.Bottom += SimHeight - Height;
         }

         GetWorkableRegion(Region);

         // Compute the width and the height of that area in taking care of
         // where the caption is to be displayed, knowing that we want the
         // bitmap to be displayed in its whole

         if ((FCaptionDisplay == cdBitmapLeft) || (FCaptionDisplay == cdBitmapRight))
            RegionWidth = Region.Right - Region.Left - BitmapWidth;
         else
            RegionWidth = Region.Right - Region.Left;

         if ((FCaptionDisplay == cdBitmapTop) || (FCaptionDisplay == cdBitmapBottom))
            RegionHeight = Region.Bottom - Region.Top - FBitmap->Height;
         else
            RegionHeight = Region.Bottom - Region.Top;

         // Split down the caption in several lines if necessary (i.e. the
         // caption is too wide for the area)

         TmpCaption = Caption;   // Avoids to alter the original caption

         if ((CaptionWidth <= RegionWidth) ||
             (RegionWidth <= 0) || (RegionHeight <= 0))
            // We are lucky: the caption fits within the area or the region is
            // null, we therefore just need to add the caption as it to the list

            CaptionList->Add(Caption);
         else
         {
            // We are unlucky: the caption doesn't fit within the area, we
            // therefore need to break it down

            NbToProcess = FCaptionNbLines;   // Maximum lines we can split
                                             // the caption into

            do
            {
               // We need to know if we require the last line or not (to know if
               // we have to add the "..." or not)
               // This may happen either when we reach the limit of number of
               // lines allowed or when the caption's height is going to be
               // greater than the control's height (if a height limit exists)

               LastOne = (NbToProcess == 1) ||
                         (((CaptionHeight * (CaptionList->Count + 2)) > RegionHeight) && aHeightLimit);

               // Get a caption line which fits within the width of the area
               // and add it to the list

               CaptionList->Add(GetACaptionLine(TmpCaption, RegionWidth,
                                                LastOne));

               NbToProcess--;
            } while (!LastOne && TmpCaption.Length());
         }

         // Compute the required width and the height to display the caption
         // (in several lines or not, it doesn't matter)

         CaptionWidth = 0;
         CaptionHeight = 0;

         for (CrtLine = 0; CrtLine < CaptionList->Count; CrtLine++)
         {
            CrtCaptionWidth = CanvasTextWidth(CaptionList->Strings[CrtLine]);

            CaptionWidth = (CrtCaptionWidth > CaptionWidth)?
                              CrtCaptionWidth:
                              CaptionWidth;

            CaptionHeight += CanvasTextHeight(CaptionList->Strings[CrtLine]);
         }
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// OptimalSize: finds out what the optimal size is for the control

void __fastcall TTransBitmap::OptimalSize(int &aWidth, int &aHeight)
{
   int Delta = 0;   // Nb of units to add to left, right, top and bottom to get
                    // the optimal size

   // The focus

   if (FAllowFocus)
      Delta += FFocusWidth;

   // Our reprensentation of the control

   if (FLook == loButton)
      // Button look: 2 frames to consider

      Delta += 2;
   else
   {
      // Panel look: inner, border and outer are to be considered

      if (FBevelInner != bvNone)
         Delta += FBevelWidth;

      Delta += FBorderWidth;

      if (FBevelOuter != bvNone)
         Delta += FBevelWidth;
   }

   // Default optimal size

   GetMaxima(aWidth, aHeight);   // Gets the biggest width & height between
                                 // the caption and the bitmap

   aWidth += 2 * Delta;    // Consider the 3D and our representation of the
   aHeight += 2 * Delta;   // control

   // Update the optimal size in case the caption has to be displayed according
   // to the bitmap

   if ((FCaptionDisplay == cdBitmapLeft) || (FCaptionDisplay == cdBitmapRight))
      aWidth = BitmapWidth + CaptionWidth + 2 * Delta;
   else if ((FCaptionDisplay == cdBitmapTop) || (FCaptionDisplay == cdBitmapBottom))
      aHeight = FBitmap->Height + CaptionHeight + 2 * Delta;

   // Depending on what has to be autosized we slightly update the optimal size

   if (FAutoSize)
   {
      if (FAutoSizeParams == asWidthOnly)
         aHeight = Height;
      else if (FAutoSizeParams == asHeightOnly)
         aWidth = Width;
   }
}

////////////////////////////////////////////////////////////////////////////////
// CheckSize: checks if we have to force the control to have a particular size
//            or not (depending on the value of "FAutoSize")

void __fastcall TTransBitmap::CheckSize(const bool &aSimulation)
{
   GetBitmapWidth();
   ComputeCaptionSize(!FAutoSize || (FAutoSizeParams != asHeightOnly), aSimulation);
   // Note: the first parameter is concerned with using a height limit or not.
   //       The height limit is to be avoided if we want to only autosize the
   //       height

   if (FAutoSize)
   {
      // Note: to be sure that the display is updated properly, we have to use
      // local variables instead of using directly "Width" and "Height" in the
      // call to "OptimalSize". This avoids to redraw in case it's not necessary

      int OptimalWidth, OptimalHeight;

      OptimalSize(OptimalWidth, OptimalHeight);

      if (aSimulation)
      {
         // That's a simulation, therefore we just update the data

         SimWidth  = OptimalWidth;
         SimHeight = OptimalHeight;
      }
      else if ((Width != OptimalWidth) || (Height != OptimalHeight))
      {
         // Not a simulation and moreover we need to update the size

         Width  = OptimalWidth;
         Height = OptimalHeight;
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// GetMaxima: finds out the maximum width and height between the caption and the
//            bitmap (very useful to align one % the other one)

void __fastcall TTransBitmap::GetMaxima(int &aMaxWidth, int &aMaxHeight)
{
   aMaxWidth = (BitmapWidth > CaptionWidth)?BitmapWidth:CaptionWidth;

   aMaxHeight = (FBitmap->Height > CaptionHeight)?FBitmap->Height:CaptionHeight;
}

////////////////////////////////////////////////////////////////////////////////
// GetBitmapWidth: computes the real width of the bitmap, this depends on the
//                 number of available picture withing the bitmap

void __fastcall TTransBitmap::GetBitmapWidth()
{
   BitmapWidth = FBitmap->Width / ((int) FBitmapNumber + 1);
}

////////////////////////////////////////////////////////////////////////////////
// GetBitmapPos: finds out where the bitmap has to be located

void __fastcall TTransBitmap::GetBitmapPos(const TRect &aRegion, TPoint &aBmpPt)
{
   int MaxWidth, MaxHeight;
   int ClientWidth = aRegion.Right - aRegion.Left;
   int ClientHeight = aRegion.Bottom - aRegion.Top;

   // Useful data for computing the position when connected to the Caption

   GetMaxima(MaxWidth, MaxHeight);

   if ((FCaptionDisplay == cdBitmapLeft) || (FCaptionDisplay == cdBitmapRight) ||
       (FCaptionDisplay == cdBitmapTop)  || (FCaptionDisplay == cdBitmapBottom))
      // Crazy case: we do have to care of the caption

      switch (FCaptionDisplay)
      {
         case cdBitmapLeft:
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aBmpPt.x = aRegion.Left; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aBmpPt.x = aRegion.Left + (ClientWidth - BitmapWidth - CaptionWidth) / 2; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aBmpPt.x = aRegion.Right - BitmapWidth - CaptionWidth; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aBmpPt.y = aRegion.Top + (MaxHeight - FBitmap->Height) / 2; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aBmpPt.y = aRegion.Top + (ClientHeight - FBitmap->Height) / 2; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aBmpPt.y = aRegion.Bottom - (MaxHeight + FBitmap->Height) / 2; break;
            }

            break;
         case cdBitmapRight:
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aBmpPt.x = aRegion.Left + CaptionWidth; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aBmpPt.x = aRegion.Left + (ClientWidth - BitmapWidth - CaptionWidth) / 2 + CaptionWidth; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aBmpPt.x = aRegion.Right - BitmapWidth; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aBmpPt.y = aRegion.Top + (MaxHeight - FBitmap->Height) / 2; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aBmpPt.y = aRegion.Top + (ClientHeight - FBitmap->Height) / 2; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aBmpPt.y = aRegion.Bottom - (MaxHeight + FBitmap->Height) / 2; break;
            }

            break;
         case cdBitmapTop:
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aBmpPt.x = aRegion.Left + (MaxWidth - BitmapWidth) / 2; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aBmpPt.x = aRegion.Left + (ClientWidth - BitmapWidth) / 2; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aBmpPt.x = aRegion.Right - (MaxWidth + BitmapWidth) / 2; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aBmpPt.y = aRegion.Top; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aBmpPt.y = aRegion.Top + (ClientHeight - FBitmap->Height - CaptionHeight) / 2; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aBmpPt.y = aRegion.Bottom - FBitmap->Height - CaptionHeight; break;
            }

            break;
         case cdBitmapBottom:
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aBmpPt.x = aRegion.Left + (MaxWidth - BitmapWidth) / 2; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aBmpPt.x = aRegion.Left + (ClientWidth - BitmapWidth) / 2; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aBmpPt.x = aRegion.Right - (MaxWidth + BitmapWidth) / 2; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aBmpPt.y = aRegion.Top + CaptionHeight; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aBmpPt.y = aRegion.Top + (ClientHeight - FBitmap->Height - CaptionHeight) / 2 + CaptionHeight; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aBmpPt.y = aRegion.Bottom - FBitmap->Height; break;
            }

            break;
      }
   else
      // Easy case: we don't care of the caption

      switch (FBitmapDisplay)
      {
         case bdTopLeft:
            aBmpPt = Point(aRegion.Left, aRegion.Top); break;
         case bdTopCenter:
            aBmpPt = Point(aRegion.Left + (ClientWidth - BitmapWidth) / 2, aRegion.Top); break;
         case bdTopRight:
            aBmpPt = Point(aRegion.Right - BitmapWidth, aRegion.Top); break;
         case bdMiddleLeft:
            aBmpPt = Point(aRegion.Left, aRegion.Top + (ClientHeight - FBitmap->Height) / 2); break;
         case bdMiddleCenter:
            aBmpPt = Point(aRegion.Left + (ClientWidth - BitmapWidth) / 2, aRegion.Top + (ClientHeight - FBitmap->Height) / 2); break;
         case bdMiddleRight:
            aBmpPt = Point(aRegion.Right - BitmapWidth, aRegion.Top + (ClientHeight - FBitmap->Height) / 2); break;
         case bdBottomLeft:
            aBmpPt = Point(aRegion.Left, aRegion.Bottom - FBitmap->Height); break;
         case bdBottomCenter:
            aBmpPt = Point(aRegion.Left + (ClientWidth - BitmapWidth) / 2, aRegion.Bottom - FBitmap->Height); break;
         case bdBottomRight:
            aBmpPt = Point(aRegion.Right - BitmapWidth, aRegion.Bottom - FBitmap->Height); break;
      }
}

////////////////////////////////////////////////////////////////////////////////
// GetCaptionPos: finds out the position at which the first line of the caption
//                should be displayed, and from that we can display the other
//                lines if required

void __fastcall TTransBitmap::GetCaptionPos(const TRect &aRegion, TPoint &aTxtPt)
{
   int MaxWidth, MaxHeight;
   int ClientWidth = aRegion.Right - aRegion.Left;
   int ClientHeight = aRegion.Bottom - aRegion.Top;

   GetMaxima(MaxWidth, MaxHeight);

   // Get the Caption position according to the chosen display
   // The 4 first cases are crazy ones, since the caption must be displayed in
   // connection with the bitmap

   switch (FCaptionDisplay)
   {
      case cdBitmapLeft:
         if (FStretch)
            aTxtPt = Point(aRegion.Right - CaptionWidth, aRegion.Top + (ClientHeight - CaptionHeight) / 2);
         else
         {
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aTxtPt.x = aRegion.Left + BitmapWidth; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aTxtPt.x = aRegion.Left + (ClientWidth - BitmapWidth - CaptionWidth) / 2 + BitmapWidth; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aTxtPt.x = aRegion.Right - CaptionWidth; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aTxtPt.y = aRegion.Top + (MaxHeight - CaptionHeight) / 2; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aTxtPt.y = aRegion.Top + (ClientHeight - CaptionHeight) / 2; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aTxtPt.y = aRegion.Bottom - (MaxHeight + CaptionHeight) / 2; break;
            }
         }

         break;
      case cdBitmapRight:
         if (FStretch)
            aTxtPt = Point(aRegion.Left, aRegion.Top + (ClientHeight - CaptionHeight) / 2);
         else
         {
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aTxtPt.x = aRegion.Left; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aTxtPt.x = aRegion.Left + (ClientWidth - BitmapWidth - CaptionWidth) / 2; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aTxtPt.x = aRegion.Right - BitmapWidth - CaptionWidth; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aTxtPt.y = aRegion.Top + (MaxHeight - CaptionHeight) / 2; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aTxtPt.y = aRegion.Top + (ClientHeight - CaptionHeight) / 2; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aTxtPt.y = aRegion.Bottom - (MaxHeight + CaptionHeight) / 2; break;
            }
         }

         break;
      case cdBitmapTop:
         if (FStretch)
            aTxtPt = Point(aRegion.Left + (ClientWidth - CaptionWidth) / 2, aRegion.Bottom - CaptionHeight);
         else
         {
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aTxtPt.x = aRegion.Left + (MaxWidth - CaptionWidth) / 2; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aTxtPt.x = aRegion.Left + (ClientWidth - CaptionWidth) / 2; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aTxtPt.x = aRegion.Right - (MaxWidth + CaptionWidth) / 2; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aTxtPt.y = aRegion.Top + FBitmap->Height; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aTxtPt.y = aRegion.Top + (ClientHeight - FBitmap->Height - CaptionHeight) / 2 + FBitmap->Height; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aTxtPt.y = aRegion.Bottom - CaptionHeight; break;
            }
         }

         break;
      case cdBitmapBottom:
         if (FStretch)
            aTxtPt = Point(aRegion.Left + (ClientWidth - CaptionWidth) / 2, aRegion.Top);
         else
         {
            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdMiddleLeft: case bdBottomLeft:
                  aTxtPt.x = aRegion.Left + (MaxWidth - CaptionWidth) / 2; break;
               case bdTopCenter: case bdMiddleCenter: case bdBottomCenter:
                  aTxtPt.x = aRegion.Left + (ClientWidth - CaptionWidth) / 2; break;
               case bdTopRight: case bdMiddleRight: case bdBottomRight:
                  aTxtPt.x = aRegion.Right - (MaxWidth + CaptionWidth) / 2; break;
            }

            switch (FBitmapDisplay)
            {
               case bdTopLeft: case bdTopCenter: case bdTopRight:
                  aTxtPt.y = aRegion.Top; break;
               case bdMiddleLeft: case bdMiddleCenter: case bdMiddleRight:
                  aTxtPt.y = aRegion.Top + (ClientHeight - FBitmap->Height - CaptionHeight) / 2; break;
               case bdBottomLeft: case bdBottomCenter: case bdBottomRight:
                  aTxtPt.y = aRegion.Bottom - FBitmap->Height - CaptionHeight; break;
            }
         }

         break;
      case cdTopLeft:
         aTxtPt = Point(aRegion.Left, aRegion.Top); break;
      case cdTopCenter:
         aTxtPt = Point(aRegion.Left + (ClientWidth - CaptionWidth) / 2, aRegion.Top); break;
      case cdTopRight:
         aTxtPt = Point(aRegion.Right - CaptionWidth, aRegion.Top); break;
      case cdMiddleLeft:
         aTxtPt = Point(aRegion.Left, aRegion.Top + (ClientHeight - CaptionHeight) / 2); break;
      case cdMiddleCenter:
         aTxtPt = Point(aRegion.Left + (ClientWidth - CaptionWidth) / 2, aRegion.Top + (ClientHeight - CaptionHeight) / 2); break;
      case cdMiddleRight:
         aTxtPt = Point(aRegion.Right - CaptionWidth, aRegion.Top + (ClientHeight - CaptionHeight) / 2); break;
      case cdBottomLeft:
         aTxtPt = Point(aRegion.Left, aRegion.Bottom - CaptionHeight); break;
      case cdBottomCenter:
         aTxtPt = Point(aRegion.Left + (ClientWidth - CaptionWidth) / 2, aRegion.Bottom - CaptionHeight); break;
      case cdBottomRight:
         aTxtPt = Point(aRegion.Right - CaptionWidth, aRegion.Bottom - CaptionHeight); break;
   }
}

////////////////////////////////////////////////////////////////////////////////
// CanvasTextOut: allows to output a text containing sub/super characters

void __fastcall TTransBitmap::CanvasTextOut(int &aX, int &aY,
                                            const String &aText)
{
   if (FAllowSubAndSuperChars)
   {
      int CrtMode = 0;   // 0 = normal, 1 = sub and 2 = super
      int aTextLength = aText.Length();
      int DefFontSize = Canvas->Font->Size;
      String CharToTest;
      int FontShift, YShift = 0;

      // Adjust Y if there are some super characters to be written

      if (HasSuperChars(aText))
         aY += SuperHeight + FSuperCharDist - NormalHeight / 2;

      for (int I = 1; I < aTextLength; I++)
      {
         CharToTest = aText.SubString(I, 1);

         if (CharToTest == FEscapeChar)
         {
            // That's the escape character, therefore we just skip it
            // Note: the following one is also to be skipped, but it's done
            //       automatically done at the end of the loop

            CharToTest = aText.SubString(I + 1, 1);

            I ++;
         }
         else if (CharToTest == FSubChar)   // Switch to sub/normal mode
         {
            CrtMode = (CrtMode == 1)?0:1;

            CharToTest = "";
         }
         else if (CharToTest == FSuperChar)   // Switch to super/normal mode
         {
            CrtMode = (CrtMode == 2)?0:2;

            CharToTest = "";
         }

         switch (CrtMode)
         {
            case 0:   // Normal
               FontShift = YShift = 0;

               break;
            case 1:   // Sub
               FontShift = -FSubFontSizeDiff;
               YShift = NormalHeight + FSubCharDist - NormalHeight / 2;

               break;
            case 2:   // Super
               FontShift = -FSuperFontSizeDiff;
               YShift = -SuperHeight - FSuperCharDist + NormalHeight / 2;

               break;
         }

         Canvas->Font->Size = DefFontSize + FontShift;
         Canvas->TextOut(aX, aY + YShift, CharToTest);

         aX += Canvas->TextWidth(CharToTest);
      } 

      // Last character to consider...

      CharToTest = aText.SubString(aTextLength, 1);

      if ((CharToTest != FEscapeChar) &&
          (CharToTest != FSubChar) &&
          (CharToTest != FSuperChar))
         Canvas->TextOut(aX, aY + YShift, CharToTest);

      Canvas->Font->Size = DefFontSize;
   }
   else
      Canvas->TextOut(aX, aY, aText);
}

////////////////////////////////////////////////////////////////////////////////
// CanvasTextWidth: allows to get the width of a text containing sub/super
//                  characters

int __fastcall TTransBitmap::CanvasTextWidth(const String &aText)
{
   int Result = Canvas->TextWidth(aText);

   if (FAllowSubAndSuperChars)
   {
      String CharToTest;
      int FontShift;
      int CrtMode = 0;   // 0 = normal, 1 = sub and 2 = super
      int aTextLength = aText.Length();
      int DefFontSize = Canvas->Font->Size;;

      Result = 0;

      for (int I = 1; I < aTextLength; I++)
      {
         CharToTest = aText.SubString(I, 1);

         if (CharToTest == FEscapeChar)
         {
            // That's the escape character, therefore we just skip it
            // Note: the following one is also to be skipped, but it's done
            //       automatically done at the end of the loop

            CharToTest = aText.SubString(I + 1, 1);

            I ++;
         }
         else if (CharToTest == FSubChar)   // Switch to sub/normal mode
         {
            CrtMode = (CrtMode == 1)?0:1;

            CharToTest = "";
         }
         else if (CharToTest == FSuperChar)   // Switch to super/normal mode
         {
            CrtMode = (CrtMode == 2)?0:2;

            CharToTest = "";
         }

         switch (CrtMode)
         {
            case 0:   // Normal
               FontShift = 0; break;
            case 1:   // Sub
               FontShift = -FSubFontSizeDiff; break;
            case 2:   // Super
               FontShift = -FSuperFontSizeDiff; break;
         }

         Canvas->Font->Size = DefFontSize + FontShift;

         Result += Canvas->TextWidth(CharToTest);
      }

      // Last character to consider...

      CharToTest = aText.SubString(aTextLength, 1);

      if ((CharToTest != FEscapeChar) &&
          (CharToTest != FSubChar) &&
          (CharToTest != FSuperChar))
         Result += Canvas->TextWidth(CharToTest);

      Canvas->Font->Size = DefFontSize;
   }

   return Result;
}

////////////////////////////////////////////////////////////////////////////////
// CanvasTextHeight: allows to get the height of a text containing sub/super
//                   characters

int __fastcall TTransBitmap::CanvasTextHeight(const String &aText)
{
   int Result = Canvas->TextHeight(aText);

   if (FAllowSubAndSuperChars)
   {
      // Compute the caption line's height

      bool HasSubs   = HasSubChars(aText);
      bool HasSupers = HasSuperChars(aText);

      Result = (HasSubs || HasSupers || aText.Length())?NormalHeight:0;
      // Default size

      if (HasSubs)
         Result += SubHeight + FSubCharDist - NormalHeight / 2;

      if (HasSupers)
         Result += SuperHeight + FSuperCharDist - NormalHeight / 2;
   }

   return Result;
}

////////////////////////////////////////////////////////////////////////////////
// HasSubOrSuperChars: finds out if the text contains sub or super cars or not

bool __fastcall TTransBitmap::HasSubOrSuperChars(const String &aText,
                                                 const bool &CheckForSubChars,
                                                 const bool &CheckForSuperChars)
{
   bool Result = false;
   int aTextLength = aText.Length();
   String CharToTest;

   for (int I = 1; (I <= aTextLength) && !Result; I++)
   {
      CharToTest = aText.SubString(I, 1);

      if (CharToTest == FEscapeChar)
         // That's the escape character, therefore we just skip it
         // Note: the following one is also to be skipped, but it's done
         //       automatically at the end of the loop

         I ++;
      else if (CharToTest == FSubChar)
         Result = true && CheckForSubChars;
      else if (CharToTest == FSuperChar)
         Result = true && CheckForSuperChars;
   }

   return Result;
}

////////////////////////////////////////////////////////////////////////////////
// HasSubChars: finds out if the text contains sub characters or not

bool __fastcall TTransBitmap::HasSubChars(const String &aText)
{
   return HasSubOrSuperChars(aText, true, false);
}

////////////////////////////////////////////////////////////////////////////////
// HasSuperChars: finds out if the text contains super characters or not

bool __fastcall TTransBitmap::HasSuperChars(const String &aText)
{
   return HasSubOrSuperChars(aText, false, true);
}

////////////////////////////////////////////////////////////////////////////////
// UpdateRectSize: updates the size of a rectangle

void __fastcall TTransBitmap::UpdateRectSize(TRect &aRegion, const int aDiff)
{
   aRegion.Top    -= aDiff; aRegion.Left  -= aDiff;
   aRegion.Bottom += aDiff; aRegion.Right += aDiff;
}

////////////////////////////////////////////////////////////////////////////////
//
// MISCELLANEOUS
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// SimulateWidth: returns the width that the control would have if we changed
//                its width (useful if we don't want to repaint the control and
//                if the autosize is enabled for the height only)

int __fastcall TTransBitmap::SimulateNewWidth(const int &aNewWidth)
{
   SimWidth  = aNewWidth;
   SimHeight = Height;

   CheckSize(true);

   return SimHeight;
}

////////////////////////////////////////////////////////////////////////////////
// SimulateHeight: returns the height that the control would have if we changed
//                 its height (useful if we don't want to repaint the control
//                 and if the autosize is enabled for the width only)

int __fastcall TTransBitmap::SimulateNewHeight(const int &aNewHeight)
{
   SimWidth  = Width;
   SimHeight = aNewHeight;

   CheckSize(true);

   return SimWidth;
}

////////////////////////////////////////////////////////////////////////////////
// ForceRefresh: forces the control to refresh itself, even if the refresh is
//               not allowed

void __fastcall TTransBitmap::ForceRefresh()
{
   bool OldFAutoRefresh = FAutoRefresh;
   // Note: we use a backup, cuz one can use "ForceRefresh" even if the refresh
   //       is allowed...

   FAutoRefresh = true;

   Repaint();

   FAutoRefresh = OldFAutoRefresh;
}

////////////////////////////////////////////////////////////////////////////////
// Registration

namespace Transbitmap
{
   void __fastcall Register()
   {
      TComponentClass classes[1] = {__classid(TTransBitmap)};

      RegisterComponents("GRY", classes, 0);
   }
}

// End of file
////////////////////////////////////////////////////////////////////////////////

